/* @(#) u_clip2d.c  version:1.12  dated:84/11/20 */
/*
 * C-GKS
 *
 * COPYRIGHT 1984 - CENTRUM VOOR WISKUNDE EN INFORMATICA
 */

/*
 * SCCS file:	SCCS/s.u_clip2d.c
 * SCCS id :	u_clip2d.c 1.12
 * Last edited:	10/30/84 10:45:06
 *
 * Author:	Paul ten Hagen, Behr de Ruiter et all.
 *         	Centrum voor Wiskunde en Informatica
 *         	Amsterdam
 */

/*
 * This file contains utility functions concerning
 * clipping:
 *	st_clipw()
 * Int 	clip2d()
 * Bool ic_in_drect()
 */

#define GKS 2
#define GKS2
#define GKS1

#include "cgksincl.h"

extern	Gks	*_gkss;

/*
 * Calculate and install current clipping 
 * rectangle (in device coordinates)
 * in workstation internal statelist.
 * The ratio is preserved.
 */
st_clipw(ws)
Wss	*ws;
{
	Drect	*clipw = &ws->ws_wsis->wi_clipr;
	Nrect	nclipw, *nclipwp;

	if (_gkss->gk_clip == TRUE) {
		nclipwp = &nclipw;
		Concat(nclipwp, &ws->ws_curw, &_gkss->gk_curnt->nt_view)
	}
	else	nclipwp = &ws->ws_curw;

	ws_transf(ws, &nclipwp->n_ll, &clipw->d_ll);
	ws_transf(ws, &nclipwp->n_ur, &clipw->d_ur);
}




/*
 * clip2d() clips line from pstart to pend against pwind.
 * New, clipped, coordinates are stored in pstart and pend.
 * Returned value of 0 means line outside clipping rectangle.
 * Returned value of 1 means beginpoint outside clipping rectangle.
 * Returned value of 2 means endpoint outside clipping rectangle.
 * Returned value of 3 means begin- and endpoint outside rectangle.
 * Returned value of 4 means line inside  clipping rectangle.
 *
 * A line through P and S can be written in vector notation as:
 * R = P + lamba * (S - P)
 * For the points on the line between P and S counts:
 * 0.0 <= lamba <= 1.0
 *
 * When the edges of a rectangle are drawn to infinity
 * the cutting points of the line with the four edges of a rectangle are:
 * cutting point with line through ll and lr: P + l1 * (P - Q)
 * cutting point with line through ul and ur: P + l2 * (P - Q)
 * cutting point with line through ll and ul: P + l3 * (P - Q)
 * cutting point with line through lr and ur: P + l4 * (P - Q)
 * If a lamba is equal to 0.0 (or 1.0),
 * the cutting point falls together with P (or S),
 * and if lamba < 0.0 or lamba > 1.0 the cutting point
 * falls outside the linepiece P to S.
 *
 * Special care is needed in the following cases:
 * 1. The height or width of the clipping box is (nearly) 0.0.
 * 2. The line is (nearly) parallel to one of the edges of the box.  
 *
 */
Int
clip2d (pstart, pend, pwind)
Dc    *pstart, *pend;
Drect *pwind;
{
	register Real l1, l2, l3, l4, l;
	register Real xll, yll, xur, yur;
	Real 	xp, yp, xs, ys;
	Real 	hx, vy, dh, dv;
	Real	eps = 1.0e-6;
	/*
	 * to make calculations easier
	 * first change all coordinates from current coordinate system 
	 * to coordinate system with origin in pstart 
	 */
	xp = pstart->d_x;
	yp = pstart->d_y;
	xs = pend->d_x - xp;
	ys = pend->d_y - yp;

	/*
	 * (xll, yll) and (xur, yur) form cliprectangle.
	 * hx is width  of cliprectangle
	 * vy is height of cliprectangle:
	 */
	xll = pwind->d_ll.d_x - xp;
	yll = pwind->d_ll.d_y - yp;
	xur = pwind->d_ur.d_x - xp;
	yur = pwind->d_ur.d_y - yp;
	hx = xur - xll;
	vy = yur - yll;

	/*
	 * Now calculate cutting points of line through pstart and pend,
	 * and two horizontal boxedgelines.
	 * this results in l1 and l2 as indicated above.
	 *
	 * An adaption is made in case the clipbox has no width,
	 * or the linepiece is parallel to the horizontal boxedges:
	 */
	if ((dh = ys * hx) == 0.0) {
		dh = eps;
	}
	l1 = (yll * hx) / dh; /* =~ yll / ys  */
	l2 = (yur * hx) / dh;

	/*
	 * in the special case the line falls together with one of
	 * the horizontal boxedges, then set l2 and l1 to 1.0 and 0.0
	 */
	if (ys == 0.0 && (yur == 0.0 || yll == 0.0)) {
		l2 = 1.0;
		l1 = 0.0;
	}

	/*
	 * put the largest cutting lamba in l2, and the smallest in l1:
	 */
	if (l1 > l2) {
		l = l1; l1 = l2; l2 = l;
	}

	/*
	 * If both l1 and l2 are < 0.0 (or > 1.0) then line from
	 * P to S falls, as far as it concerns the y-coordinates,
	 * below (or above) the cliprectangles.
	 * The line is then completely clipped away:
	 */
	if ((l2 < 0) || (l1 > 1))
		return(0);

	/*
	 * Now do the same for the left and right vertical sides
	 * of the cliprectangle.
	 * This results in l3 and l4, with l4 >= l3.
	 */
	if ((dv = xs * vy) == 0.0) {
		dv = eps;
	}
	l3 = (xll * vy ) / dv;
	l4 = (xur * vy ) / dv;

	if (xs == 0.0 && (xur == 0.0 || xll == 0.0)) {
		l4 = 1.0;
		l3 = 0.0;
	}
	if (l3 > l4) {
		l = l3; l3 = l4; l4 = l;
	}
	/*
	 * if the line is completely to the right (l4 >= l3 > 1.0)
	 * or to the left ( l3 <= l4 < 0.0)
	 * then line outside cliprectangle:
	 */
	if ((l3 > 1) || (l4 < 0))
		return(0);

	/*
	 * The resulting, clipped line is now the intersection of the
	 * following three linepieces:
	 * linepiece from P to S
	 * linepiece from P + l1 * (P - S) to P + l2 * ( P - S) 
	 * linepiece from P + l3 * (P - S) to P + l4 * ( P - S) 
	 *
	 * if the intersection of the last two pieces is empty
	 * then the line is clipped away too:
	 */
	if ((l2 < l3) || (l4 < l1))
		return(0);

	/*
 	 * we now have covered all outside cases.
 	 * select candidate cutting points in l1 and l2
 	 */

	if (l1 < l3) l1 = l3;
	if (l2 > l4) l2 = l4;

	if (l1 <= 0) {
		if (l2 >= 1)	return(4);
		else {	pend->d_x = xp + l2 * xs;
			pend->d_y = yp + l2 * ys;
			return(2);
		}
	}
	else {
		pstart->d_x = xp + l1 * xs;
		pstart->d_y = yp + l1 * ys;
		if (l2 < 1) {
			pend->d_x = xp + l2*xs;
			pend->d_y = yp + l2*ys;
			return(3);
		}
		else	return(1);
	}
}

Bool
ic_in_drect(drect, ic)
Drect	*drect;
Ic	*ic;
{	return  (Bool) ((ic->i_x >= drect->d_ll.d_x ) &&
		    (ic->i_y >= drect->d_ll.d_y ) &&
		    (ic->i_x <= drect->d_ur.d_x ) &&
		    (ic->i_y <= drect->d_ur.d_y )
		   ); 
}

