/* @(#) u_matrix.c  version:1.9  dated:84/11/20 */
/*
 * C-GKS
 *
 * COPYRIGHT 1984 - CENTRUM VOOR WISKUNDE EN INFORMATICA
 */

/*
 * SCCS file:	SCCS/s.u_matrix.c
 * SCCS id :	u_matrix.c 1.9
 * Last edited:	10/30/84 15:29:50
 *
 * Author:	Behr de Ruiter, Paul ten Hagen
 *         	Centrum voor Wiskunde en Informatica
 *         	Amsterdam
 */

/*
 * This file contains the following utility functions for 
 * transformation functions and the like used everywhere in the system.
 *	concat();
 *	cp_mat();
 * Bool	dc_in_drect();
 *	dc_to_ic();
 *	invers_mat();
 *	matmat();
 *	ntransf();
 *	s_idmatrix();
 *	scale();
 *	set_mat();
 *	setwsmat();
 * Bool	unit_mat();
 * Bool	unit_rect();
 *	vecmat();
 *	vecmat_restore();
 *	vecrotate();
 *	ws_transf();
 */

#define GKS 2
#define GKS2
#define GKS1

#include "cgksincl.h"

extern	Gksi	*gksis;

/*
 * cp_mat copies oldmat(6 Reals) to matprt 
 * in fact both are Tmat pointers.
 */
cp_mat(oldmat, matptr)
Real	*oldmat, *matptr;
{	Int i;
	for (i = 0; i < 6 ; i++) *matptr++ = *oldmat++;
}

/*
 * vecrotate calculates vector times matrix, without
 * interpreting translation.
 * vecnew = vec * mat
 */
vecrotate(vecnew, vec, mat)
Real *vecnew, *vec, *mat;
{	*vecnew = *vec       * *mat +
		  *(vec + 1) * *(mat + 1);
	vecnew++;
	*vecnew = *vec       * *(mat + 3) +
		  *(vec + 1) * *(mat + 4);
}

/*
 * vecmat calculates vector times tranformation matrix.
 * vecnew = vec * mat
 */
vecmat(vecnew, vec, mat)
Real *vecnew, *vec, *mat;
{	*vecnew = *vec       * *mat +
		  *(vec + 1) * *(mat + 1) +
	                       *(mat + 2);
	vecnew++;
	*vecnew = *vec       * *(mat + 3) +
		  *(vec + 1) * *(mat + 4) +
	                       *(mat + 5);
}

/* vecmat_restore calculates vec * mat , and restores it
 * in vec.
 */ 
vecmat_restore(vec, mat)
Real	*vec, *mat;
{	Real vecnew[2];
	vecnew[0] = *vec       * *mat +
		    *(vec + 1) * *(mat + 1) +
	                         *(mat + 2);
	vecnew[1] = *vec       * *(mat + 3) +
		    *(vec + 1) * *(mat + 4) +
	                         *(mat + 5);
	*vec = vecnew[0];
	vec++;
	*vec = vecnew[1];
}

/* 
 * matmat calculates new Tmat tm3 by muliplying Tmat tm and tm2:
 * tm3 = tm * tm2 
 */
matmat(tm, tm2, tmres)
Tmat tm, tm2;
Real *tmres;
{
	*(tmres++) = tm[0][0]*tm2[0][0] + tm[0][1]*tm2[1][0]; 
	*(tmres++) = tm[0][0]*tm2[0][1] + tm[0][1]*tm2[1][1]; 
	*(tmres++) = tm[0][0]*tm2[0][2] + tm[0][1]*tm2[1][2] + tm[0][2];
	*(tmres++) = tm[1][0]*tm2[0][0] + tm[1][1]*tm2[1][0]; 
	*(tmres++) = tm[1][0]*tm2[0][1] + tm[1][1]*tm2[1][1]; 
	*tmres     = tm[1][0]*tm2[0][2] + tm[1][1]*tm2[1][2] + tm[1][2];
}

/*
 * set_mat calculates out of a window *wind (a pointer to a
 * lowerleft x and y, and upperright x and y), and a viewport *view
 * a transformation matrix.
 * If the Bool prsrv is TRUE the viewport is clipped first, i.e.
 * either a slice at the right side or at the above side
 * is taken off to make the ratio of both rectangles equals,
 * so the ratio is preserved.
 */
set_mat(wind, view, mat, prsrv)
Real	*wind, *view;
register Real *mat;
Bool	prsrv;
{
	Real ux, vy;

	ux = (*(view + 2) - *view      ) / (*(wind + 2) - *wind      );
	vy = (*(view + 3) - *(view + 1)) / (*(wind + 3) - *(wind + 1));
	if (prsrv == TRUE)  {
		if (vy > ux)	vy = ux;
		else		ux = vy;
	}
	*mat++ = ux;
	*mat++ = 0.0;
	*mat++ = *view       - ux * *wind;
	*mat++ = 0.0;
	*mat++ = vy;
	*mat++ = *(view + 1) - vy * *(wind + 1);
}

/*
 * setwsmat sets the workstation transformation matrix in the
 * workstation internal statelist, calculated from the
 * current window and viewport.
 */
setwsmat(ws)
Wss	*ws;
{
	set_mat((Real *)(&ws->ws_curw), (Real *)(&ws->ws_curv),
		&ws->ws_wsis->wi_mat[0][0], TRUE);
}

/*
 * scale WC coordinate  wht to NDC coordinate nht according to
 * transformation matrix mat.
 * (no translation.)
 */	
scale(mat, wht, nht)
Tmat	mat;
Wc	*wht;
Nc	*nht;
{	
	nht->n_x = mat[0][0] * wht->w_x + mat[0][1] * wht->w_y ;
	nht->n_y = mat[1][0] * wht->w_x + mat[1][1] * wht->w_y ;
}

/*
 * transform NDC to DC via workstation transformation of ws
 */
ws_transf(ws, nc, dc)
Wss	*ws;
Nc	*nc;
Dc	*dc;
{	vecmat((Real *)dc, (Real *)nc, &ws->ws_wsis->wi_mat[0][0]);
}

/*
 * transform Wc coordinates to NDC coordinates with 
 * current normalization transformation.
 * sufficient to do only scale and shift (no rotate).
 */
ntransf(wc, nc)
Wc *wc;
Nc *nc;
{
/*
	vecmat((Real *)nc, (Real *)wc, &gksis->gki_mat[0][0]);
*/
	nc->n_x = wc->w_x * gksis->gki_mat[0][0] + gksis->gki_mat[0][2];
	nc->n_y = wc->w_y * gksis->gki_mat[1][1] + gksis->gki_mat[1][2];
}


/*
 * invers_mat calculates the inverse transformation matrix
 * of tmat. 
 * Result is in imat, where imat is the address of the first element
 * of a Tmat.
 */
invers_mat(tmat, imat, rout)
Tmat	tmat;
Real	*imat;
Int	rout;	/* number/identifier of calling routine */
{	register Real	det;

	det = (tmat[0][0]*tmat[1][1] - tmat[0][1]*tmat[1][0]);
	if (fabs(det) < 1.0e-6) {
		Rep_Er(309, rout);  
	}
	*imat++ =  tmat[1][1] / det;
	*imat++ = -tmat[0][1] / det;
	*imat++ = (tmat[0][1] *tmat[1][2] - tmat[0][2]*tmat[1][1])/det;
	*imat++ = -tmat[1][0] / det;
	*imat++ =  tmat[0][0] / det;
	*imat   = (tmat[0][2] *tmat[1][0] - tmat[0][0]*tmat[1][2])/det;
	return(0);
}

/*
 * intersect the inner of 2 rectangles w1 and w2.
 * result in newr, being a pointer to first of 4 Reals (ll and ur). 
 */
concat(newr, w1, w2)
Real	*newr, *w1, *w2;
{
	*newr++ = (*w1     < *w2    ) ? *w2 : *w1;
	*newr++ = (*(++w1) < *(++w2)) ? *w2 : *w1;
	*newr++ = (*(++w2) < *(++w1)) ? *w2 : *w1;
	*newr   = (*(++w2) < *(++w1)) ? *w2 : *w1;
}

/*
 * fills Tmat , pointed to by mat, with unity matrix.
 */
s_idmatrix(mat)
register Real	*mat;
{
	*mat++ = 1.0; *mat++ = 0.0; *mat++ = 0.0;
	*mat++ = 0.0; *mat++ = 1.0; *mat   = 0.0;
}

/*
 * unit_mat returns TRUE if the matrix Tmat pointed to by mat,
 * is the unity matrix.
 */
Bool
unit_mat(mat)
register Real	*mat;
{
	return((Bool) (
		    *mat == 1.0 && *(++mat) == 0.0 && *(++mat) == 0.0 &&
		*(++mat) == 0.0 && *(++mat) == 1.0 && *(++mat) == 0.0
	      )       );
}

/*
 * make an coordinate, consisting of 2 ints, out of a 
 * coordinate consisting of 2 reals.
 * This is done by truncation.
 */
dc_to_ic(ic, dc)
Ic	*ic;
Dc	*dc;
{
	ic->i_x = dc->d_x;
	ic->i_y = dc->d_y;
}

/* 
 * does coordinate dc lay in rectangle drect ?
 * dc_in_drect returns also TRUE if on boundary.
 */
Bool
dc_in_drect(drect, dc)
Drect	*drect;
Dc	*dc;
{
	return  (Bool) ((dc->d_x >= drect->d_ll.d_x ) &&
		    (dc->d_y >= drect->d_ll.d_y ) &&
		    (dc->d_x <= drect->d_ur.d_x ) &&
		    (dc->d_y <= drect->d_ur.d_y )
		   ); 
}


/* nowhere used at this moment:
Bool
unit_rect(rect)
Real	*rect;
{
	return((Bool) (
		*rect       == 0.0 && 
		*(rect + 1) == 0.0 &&
	      	*(rect + 2) == 1.0 && 
		*(rect + 3) == 1.0
	      ));
}
*/
